
#include <stdlib.h>
#include "types.h"
#include "CDib.h"
#include "nodes.h"
#include "tools.h"
#include "snapshots.h"
#include "resource.h"
#include "afxres.h"

#include <conio.h>


#define LONIBBLE(x)		(x&0x0f)
#define HINIBBLE(x)		((x&0xf0)>>4)
#define BIT0(x)			(x&0x01)
#define BIT1(x)			(x&0x02)
#define BIT2(x)			(x&0x04)
#define BIT3(x)			(x&0x08)
#define BIT4(x)			(x&0x10)
#define BIT5(x)			(x&0x20)
#define BIT6(x)			(x&0x40)
#define BIT7(x)			(x&0x80)

extern HINSTANCE	hInst;
extern HWND			hWndMain;

void DrawRectangle ( int x, int y, int w, int h, COLORREF color );
void Text ( int x, int y, LPSTR text );
void Line ( int x, int y, int x1, int y1, COLORREF color );
void CreateSpectrumPalette ( void ) ;

void DrawPlatform ( int startx, int starty, int platform, int& width, int& height );
void DisplayTableGraphics ( int x, int y, u16 table[], int total );
void DisplayGraphic ( int x, int y, u8* data, int width, int height, u8 attr );
void PlotByte ( int x, int y, u8 byte, u8 attr );
void PlotByteReverse ( int x, int y, u8 byte, u8 attr );
void DrawPlatform ( int startx, int starty, int platform );
void Draw8x8 ( int x, int y, int block );
void Draw8x9 ( int x, int y, int block );
void DisplayRoom ( u8 room );
void DisplaySprites ( void );
void DisplayPlatforms ( void );
u8* FindRoomData ( u8 room );
u8* FindBorderData ( u8 border );
void DisplayBorder ( u8 border );
void DisplaySprite ( int startx, int starty, u16 table[], int sprite );
void DisplayMovingObjects ( int room );
void DisplayObjects ( int room );
void DisplayMoles ( int room );
void DrawGrid ( void );


char szAppName[] = APPNAME; // The name of this application
char szTitle[]   = APPNAME": Sample Application"; // The title bar text

CDib*		pSpecScreen;
HPALETTE	hPalSpec;
u8*			specmem=NULL;
HDC			hDcMain;
int			room;
char number[10];


void nodes_Init()
{
	CreateSpectrumPalette();
	pSpecScreen = new CDib( GetDC(NULL), SCREEN_WIDTH, SCREEN_HEIGHT, hPalSpec );
	specmem = OpenSNAfile ( "d:\\nodes\\nodes.sna" );

	room = 0x32;
}

void nodes_DeInit()
{
	SAFEDELETE ( pSpecScreen );
	SAFEFREE ( specmem );
}


u16 ascii_sets[] = {
	0x62c0, 0x670e, 0x6b6d, 0x6eeb, 0x7387, 0x7726, 
};

u16 pixel_sets[] = {
	0x635d, 0x67aa, 0x6bf7, 0x6f85, 0x73e6, 0x781d, 
};

u16 cur_ascii;
u16 cur_pixel;


void nodes_Paint ( HDC hDC )
{
RECT	rect;

	hDcMain = hDC;

	if ( !pSpecScreen ) return;

//	SetRect ( &rect, 0, 0, SCREEN_WIDTH*2, SCREEN_HEIGHT*2 );
//	FillRect ( hDcMain, &rect, GetStockObject(BLACK_BRUSH) );
//	DisplaySprites();
//	return;

	memset ( pSpecScreen->m_pBits,0x00, SCREEN_WIDTH*SCREEN_HEIGHT );
	DisplayRoom(room);
	DisplayMovingObjects(room);
	DisplayObjects(room);
	DisplayMoles(room);
	pSpecScreen->Draw( hDcMain, 0, 0, 0, 0, 256, 192, 2 );

//	DrawGrid();
}

void DrawGrid ( void )
{

	for ( int x=0; x<512; x+=16 )
		Line ( x, 0, x, 384, RGB(255,255,255) );
	for ( int y=0; y<384; y+=16 )
		Line ( 0, y, 512, y, RGB(255,255,255) );
	sprintf ( number, "%02x", room );
	Text ( 16, 340, number );

}

void DisplayMoles ( int room )
{
	u8* data = &specmem[0xa299];
	for ( int ii=0; ii<6; ii++ ) {
		if ( *data == room ) {
			DisplaySprite ( data[1], data[2], (u16*)&specmem[0xc700], 0x32 );
			DisplaySprite ( data[3], data[4], (u16*)&specmem[0xc700], 0x32 );
		}
		data+=5;
	}
}

void DisplayMovingObjects ( int room )
{
u8* data = &specmem[0xaaed];
int x,y,gfx;

	for ( int ii=0; ii<94; ii++ ) {
		if ( data[2] == room ) {
			u16 adr = ((data[1]|0x80)*256)+data[0];
			gfx= specmem[adr]&0x7f;
			x=HINIBBLE(data[4])*16;
			y=data[5];
			DisplaySprite ( x, y, (u16*)&specmem[0xc700], gfx );
		}
		data+=8;
	}
}

void DisplayObjects ( int room )
{

	// can the room hold an object?
	if ( BIT7(specmem[0xb600+room]) == 0 )
		return;

	u8* data = &specmem[0x5d68];
	u8* rooms = &specmem[0xb600];
	// location its coordinates

	for ( int ii=0; ii<=room; ii++  ) {
		if ( BIT7(*rooms) )
			data++;
		rooms++;
	}

	//int x = LONIBBLE(*data)*16;
	//int y = (19*8)-(HINIBBLE(*data)*16);

	u8 y = (LONIBBLE(*data)*8);
	u8 x = (HINIBBLE(*data)*16);

	if ( BIT4(specmem[0xb600+room]) )
		DisplaySprite ( x, y, (u16*)&specmem[0x5b54], 4 );
	else
		DisplaySprite ( x, y, (u16*)&specmem[0x5b54], 1 );

}


void AboveGroundLevel ( int border )
{
u8* data = &specmem[0x5FA7];
u8	a,b,c,d,e;
u16	aa;
int	width,height;
int ii,x,y,gfx;

	cur_ascii=ascii_sets[3];
	cur_pixel=pixel_sets[3];

	while ( --border )
		data+=LONIBBLE(*data);

	memset ( (u8*)pSpecScreen->m_pBits+(8*8*SCREEN_WIDTH),1, SCREEN_WIDTH*10*8 );

	// draw bottom
	x = 0; 
	y = 18*8;
	d = *data++ ;
	c = d;
	for ( ii=0; ii<4; ii++ ) {
		gfx = 4 + (BIT7(c)>>7); c<<=1;
		DrawPlatform ( x, y, gfx, width, height );
		x+=64;
	}

	// draw top rocks
	x=0;
	y=6*8;

	aa=(data[0]<<8)|data[1]; data+=2;
	for ( ii=0; ii<8; ii++ ) {
		gfx = (aa&0xc000)>>14; aa<<=2;
		DrawPlatform ( x, y, gfx, width, height );
		x+=32;
	}


	// draw craters and volcanoes
	d = LONIBBLE(d)-3;
	for ( ii=0; ii<d; ii++ ) {
		c=*data++;
		x=HINIBBLE(c)*2;
		y=((c>>1)&0x07)+11; // y coor
		gfx=6+BIT0(c); c>>=1;
		DrawPlatform ( x*8, y*8, gfx, width, height );
		// if gfx is 6 then remember the exit
	}

	// draw the moon
	DrawPlatform ( 25*8, 2*8, 8, width, height );
}

void DisplayRoom ( u8 room )
{
u8* data;
u8	x;
u8	y;
u8	p;
int width,height;

	data = &specmem[0xb500];
    u8 gfxset = ((data[room]&0xc0)>>6);
    u8 border = (data[room]&0x3f)+1;

	if ( gfxset == 0 ) {
		AboveGroundLevel(border);
		return;
	}

	// set thr graphics set
	cur_ascii=ascii_sets[gfxset-1];
	cur_pixel=pixel_sets[gfxset-1];

	DisplayBorder ( border );

	data = &specmem[0xb600];
	data = FindRoomData ( room );

	// any actual platforms here
	if ( *data & 0x40 ) 
		return;

	do {
		x = (data[0])&0x1f;
		y = HINIBBLE(data[1]) + 4;
		p = LONIBBLE(data[1]);
		data+=2;

		DrawPlatform ( x*8, y*8, p, width, height );

	} while ( !BIT7(data[0]) );


}





void DisplayBorder ( u8 border )
{
u8* data;
u8 a,b,c,d;
u16	aa;
int width,height;
int ii,x,y;
int	gfx;
u8 mask;
int temp1;
int temp2;

	data = FindBorderData(border);

	// 
	d = *data++;

	// draw top border?
	if ( BIT0(d) ) {
		a = *data++;

		//
		gfx = (a&0xc0)>>6;
		DrawPlatform ( -16, 0, gfx, width, height );

		mask = LONIBBLE(a);
		mask = specmem[0x5E2C+mask];
		x=16;
		y=0;

		aa = data[0]<<8|data[1]; data+=2;
		for ( ii=0; ii<8; ii++ )  {
			gfx=(aa&0xc000)>>14; aa<<=2;
			if ( BIT7(mask) )
				DrawPlatform ( x, y, gfx, width, height );
			x+=32;
			mask<<=1;
		}


	}

	// draw left border?
	if ( BIT1(d) ) {
		c=*data++;
		x=0;
		y=0;
		for ( ii=0; ii<7; ii++ ) {
			gfx = 4 + (BIT7(c)>>7); c<<=1;
			DrawPlatform ( x, y, gfx, width, height );
			y+=24;
		}
	}

	// draw right border?
	if ( BIT2(d) ) {
		c=*data++;
		x=(30*8);
		y=0;
		for ( ii=0; ii<7; ii++ ) {
			gfx = 6 + (BIT7(c)>>7); c<<=1;
			DrawPlatform ( x, y, gfx, width, height );
			y+=24;
		}
	}

	// draw bottom border?
	if ( BIT3(d) ) {
		c=*data++;
		if ( HINIBBLE(d) == 0 ) {
			x=0;
			y=19*8;
			for ( ii=0; ii<8; ii++ ){
				gfx = 8 + (BIT7(c)>>7); c<<=1;
				DrawPlatform ( x, y, gfx, width, height );
				x+=32;
			}
		}
	}

	// draw bottom corners?
	x=0;
	y=19*8;
	mask = (d>>3)&0x1e;
	aa = specmem[0x5E35+mask]<<8|specmem[0x5E35+mask+1];
	
	for ( ii=0; ii<8; ii++ ) {
		gfx = (aa&0xc000)>>14; aa<<=2;
		if ( gfx == 1 )
			DrawPlatform ( x, y, 8 + (BIT7(c)>>7), width, height );
		else if ( gfx == 2 )
			DrawPlatform ( x, y, 0xa, width, height );
		else if ( gfx == 3 )
			DrawPlatform ( x, y, 0xb, width, height );
		x+=32;
		c<<=1;
	}

}




u8* FindBorderData ( u8 border )
{
u8* data;
u8	d;

	data = &specmem [ 0x5e4f ];

	while ( border ) {
		d = *data++;
		if ( BIT0(d) )
			data+=3;
		if ( BIT1(d) )
			data++;
		if ( BIT2(d) )
			data++;
		if ( BIT3(d) )
			data++;
		border--;
	}

	return data ;
}


u8* FindRoomData ( u8 room )
{
	int temproom = room - 16;
	u8* data = &specmem[0xdbff] ;
	while ( temproom ) {
		do { data+=2; } while ( !BIT7(*data) );
		temproom--;
	};
	return data ;	
}


void DisplaySprites ( void )
{
	DisplayTableGraphics ( 3, 12, (u16*) &specmem[0xc700], 112 );
	DisplayTableGraphics ( 3, 200, (u16*) &specmem[0xb700], 41 );
	DisplayTableGraphics ( 100, 340, (u16*) &specmem[0x5b54], 5 );
}

void DisplayPlatforms ( void )
{
int		width,height;
int		x,y;
int		highest;
	y=3;

	for ( int jj=0; jj<4; jj++ ) {
		cur_ascii=ascii_sets[jj];
		cur_pixel=pixel_sets[jj];

		x=3;
		highest=0;

		for ( int ii=0; ii<16; ii++ ) {

			DrawPlatform ( 0, 0, ii, width, height );
			pSpecScreen->Draw( hDcMain, x, y, 0, 0, width, height );

			DrawRectangle ( x-1, y-1, width+2, height+2, RGB(255,0,255) );
			//if ( width==0 || height == 0 ) break;
			if ( height > highest ) highest=height;
			x+=width+3;

			if ( x+32 > SCREEN_WIDTH ) {
				y+= highest+3;
				highest=0;
				x=3;
			}

		}

		y += highest+3;

	}

}

void DisplayNormalSprite ( int startx, int starty, u16 table[] )
{
}


void DisplaySprite ( int startx, int starty, u16 table[], int sprite )
{
int pixelwidth,bytewidth, height;
u8	attr;

	u8* data = &specmem [ table[sprite] ];
	pixelwidth = *data++;
	height = *data++;
	attr = *data++;
	bytewidth = (pixelwidth/8);
	if ( pixelwidth%8 ) bytewidth++;
	DisplayGraphic ( startx, starty, data, bytewidth, height, attr );

}




void DisplayTableGraphics ( int startx, int starty, u16 table[], int total )
{
int offset;
int pixelwidth,bytewidth, height;
int x,y;
int highest;
u8	attr;

	x= startx;
	y= starty;
	highest=0;

	int count=0;

	for ( int ii=0; ii<total; ii++ ) {

		u8* data = &specmem [ table[ii] ];

		pixelwidth = *data++;
		height = *data++;
		attr = *data++;

		bytewidth = (pixelwidth/8);
		if ( pixelwidth%8 ) bytewidth++;

		pixelwidth = bytewidth*8 ;

		if ( x+pixelwidth > SCREEN_WIDTH || count == 16 ) {
			y+= highest+3;
			highest=0;
			x=startx;
			count=0;
		}

		if ( y+height > SCREEN_HEIGHT )
			break;

		//x=count*40+3;
		DisplayGraphic ( 0, 0, data, bytewidth, height, attr );
		pSpecScreen->Draw( hDcMain, x, y, 0, 0, pixelwidth, height );
		DrawRectangle ( x-1, y-1, pixelwidth+2, height+2, RGB(255,0,255) );

		//sprintf ( number, "%02x", count );
		//Text ( x, y-6, number );

		if ( height > highest ) highest=height;
		x+= (pixelwidth)+3;
		//count++;
	}
}

void DisplayGraphic ( int x, int y, u8* data, int width, int height, u8 attr )
{
	for ( int y1=0; y1<height; y1++ ) {
		for ( int x1=0; x1<width; x1++ ) {
			PlotByte ( x+(x1*8), y+y1, *data++, attr );
		}
	}
}


void DrawPlatform ( int startx, int starty, int platform, int& width, int& height )
{
u8* data = &specmem[cur_ascii];
int offset = 0 ;
int count = platform + 1 ;
//int height = 0;
//int width = 0;

	height = 0;
	width = 0;

    for ( int ii=0; ii<platform+1; ii++ ) {
        data += width * height;
        height = LONIBBLE(*data);
        width = HINIBBLE(*data);
		data++; // skip attribute
    }
    
	for ( int y=0; y<height; y++ ) {
		for ( int x=0; x<width; x++ )
			Draw8x9 ( startx+(x*8), starty+(y*8), *data++ );
    }

	width*=8;
	height*=8;
}



void Draw8x9 ( int x, int y, int block )
{
	u8* data = &specmem[ ((block&0x7f)*9)+cur_pixel ];
	int ii;

	if ( BIT7(block) ) {
		for ( ii=0; ii<8; ii++ )
			PlotByteReverse ( x, y+ii, data[ii], data[8] );
	}else{
		for ( ii=0; ii<8; ii++ )
			PlotByte ( x, y+ii, data[ii], data[8] );
	}
}

void Draw8x8 ( int x, int y, int block )
{
	u8* data = &specmem[ (block*8)+cur_pixel ];
	for ( int ii=0; ii<8; ii++ )
		PlotByte ( x, y+ii, data[ii], 7 );
}

void PlotByte ( int x, int y, u8 byte, u8 attr )
{
	if ( x <0 || x > SCREEN_WIDTH )
		return;

u8* dst = (u8*)pSpecScreen->m_pBits;

	dst += y*SCREEN_WIDTH;
	dst += x;

    u8 ink = attr & 0x07;
    u8 pap = (attr>>3) & 0x07;
    u8 bright = ((attr>>6)&0x01)<<3 ;
    u8 flash = (attr>>7)&0x01 ;
    ink |= bright ;

	for ( int x1=0; x1<8; x1++ ) {
		if ( BIT7(byte) )
			*dst++ = ink;
			else
			*dst++ = pap;
		byte<<=1;
	}

}


void PlotByteReverse ( int x, int y, u8 byte, u8 attr )
{
	if ( x <0 || x > SCREEN_WIDTH )
		return;
	
u8* dst = (u8*)pSpecScreen->m_pBits;

	dst += y*SCREEN_WIDTH;
	dst += x;
    u8 ink = attr & 0x07;
    u8 pap = (attr>>3) & 0x07;
    u8 bright = ((attr>>6)&0x01)<<3 ;
    u8 flash = (attr>>7)&0x01 ;
    ink |= bright ;

	for ( int x1=0; x1<8; x1++ ) {
		if ( BIT0(byte) )
			*dst++ = ink;
			else
			*dst++ = pap;
		byte>>=1;
	}

}



void nodes_OnCommand(HWND hWnd, int id, HWND hWndCtl, UINT codeNotify)
{
	switch ( id ) {

		case ID_APP_EXIT:
			DestroyWindow (hWnd);
			break;

		case ID_FILE_SAVE:
			pSpecScreen->Save("d:\\test.bmp");
			break;

		case ID_RIGHT:
			if ( LONIBBLE(room) < 15 ) {
				room++;
				InvalidateRect(hWnd,NULL,FALSE);
			}
			break;
		case ID_LEFT:
			if ( LONIBBLE(room) ) {
				room--;		
				InvalidateRect(hWnd,NULL,FALSE);
			}
			break;
		case ID_UP:
			if ( room>15 ) {
				room-=16;
				InvalidateRect(hWnd,NULL,FALSE);
			}
			break;
		case ID_DOWN:
			if ( room <240 ) {
				room+=16;
				InvalidateRect(hWnd,NULL,FALSE);
			}
			break;

		default:
			FORWARD_WM_COMMAND (hWnd, id, hWndCtl, codeNotify, DefWindowProc );
	}

}



RGBQUAD rgbtable[] = {
// Normal colours
	{ 0x00, 0x00, 0x00 },    // black
	{ 0x00, 0x00, 0xcf },    // blue
	{ 0xcf, 0x00, 0x00 },    // red
	{ 0xcf, 0x00, 0xcf },    // Magenta
	{ 0x00, 0xcf, 0x00 },    // green
	{ 0x00, 0xcf, 0xcf },    // Cyan
	{ 0xcf, 0xcf, 0x00 },    // yellow
	{ 0xcf, 0xcf, 0xcf },    // white
// Bright colours
	{ 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0xff },
	{ 0xff, 0x00, 0x00 },
	{ 0xff, 0x00, 0xff },
	{ 0x00, 0xff, 0x00 },
	{ 0x00, 0xff, 0xff },
	{ 0xff, 0xff, 0x00 },
	{ 0xff, 0xff, 0xff },
};


void CreateSpectrumPalette ( void ) 
{
HANDLE		hLogPal;
LOGPALETTE*	pLogPal;

	hLogPal = GlobalAlloc( GHND, sizeof(LOGPALETTE) + sizeof(PALETTEENTRY)*256 );
	pLogPal = (LOGPALETTE*)GlobalLock( hLogPal );

	pLogPal->palVersion = 0x300;
	pLogPal->palNumEntries = 256;

	for( int i=0; i < NUMELE(rgbtable); i++ )
		memcpy ( &pLogPal->palPalEntry[i], &rgbtable[i], sizeof(RGBQUAD) );

	hPalSpec = CreatePalette( pLogPal );

	GlobalUnlock( hLogPal );
	GlobalFree( hLogPal );
}

void DrawRectangle ( int x, int y, int w, int h, COLORREF color )
{
HPEN hPen;
HPEN hOldPen;
HBRUSH hOldBrush;

	hPen = CreatePen ( PS_SOLID, 1, color );
	hOldPen = SelectObject ( hDcMain, hPen );
	hOldBrush = SelectObject ( hDcMain, GetStockObject( NULL_BRUSH ) );
	Rectangle ( hDcMain, x, y, x+w, y+h );
	SelectObject ( hDcMain, hOldBrush );
	SelectObject ( hDcMain, hOldPen );
	DeleteObject ( hPen );
}

void Line ( int x, int y, int x1, int y1, COLORREF color )
{
HPEN	hPen;
HPEN	hOldPen;
HBRUSH	hOldBrush;

	hPen = CreatePen ( PS_SOLID, 1, color );
	hOldPen = SelectObject ( hDcMain, hPen );
	hOldBrush = SelectObject ( hDcMain, GetStockObject( NULL_BRUSH ) );
	MoveToEx ( hDcMain, x, y, NULL );
	LineTo ( hDcMain, x1, y1 );
	SelectObject ( hDcMain, hOldBrush );
	SelectObject ( hDcMain, hOldPen );
	DeleteObject ( hPen );

}

void Text ( int x, int y, LPSTR text )
{
	if ( text == NULL )
		return;

	TextOut ( hDcMain, x, y, text, strlen(text) );
}